@@ -1,29 +1,35 @@
-"""Ops and optimizations: sigmoid, softplus
-
-These functions implement special cases of exp and log to improve numerical stability.
 """
+Ops and optimizations: sigmoid, softplus.
+
+These functions implement special cases of exp and log to improve numerical
+stability.
+
+"""
+from __future__ import print_function
 
 import warnings
-from itertools import imap
 
 import numpy
 
 import theano
 from theano import config, gof, printing, scalar
-from theano.configparser import AddConfigVar, BoolParam
+from theano.compat import imap
 from theano.printing import pprint
 from theano.tensor import basic as tensor
 from theano.tensor import elemwise, opt, NotScalarConstantError
-
+from theano.tensor.type import values_eq_approx_remove_inf
+from theano.tensor.opt import copy_stack_trace
 
 ############
 #
 # SCALAR OPS
 #
 
+
 class ScalarSigmoid(scalar.UnaryScalarOp):
     """
     This is just speed opt. Not for stability.
+
     """
     @staticmethod
     def st_impl(x):
@@ -31,6 +37,11 @@ class ScalarSigmoid(scalar.UnaryScalarOp):
             return 0.0
         if x > 30.0:
             return 1.0
+        # If x is an int8 or uint8, numpy.exp will compute the result in
+        # half-precision (float16), where we want float32.
+        x_dtype = str(getattr(x, 'dtype', ''))
+        if x_dtype in ('int8', 'uint8'):
+            return 1.0 / (1.0 + numpy.exp(-x, sig='f'))
         return 1.0 / (1.0 + numpy.exp(-x))
 
     def impl(self, x):
@@ -52,16 +63,21 @@ class ScalarSigmoid(scalar.UnaryScalarOp):
         # We add boundary checks prevent exp from generating inf or
         # 0. The reset of the logic always generate 0 or 1 in those
         # cases. This is a speed optimization.
-        # The constants were obtained by looking at the output of python commands like:
-        """
-import numpy, theano
-dt='float32'  # or float64
-for i in xrange(750):
-    print i, repr(theano._asarray(1.0, dtype=dt) /
-                              (theano._asarray(1.0, dtype=dt) +
-                              numpy.exp(-theano._asarray([i,-i], dtype=dt))))
-        """
-        if node.inputs[0].type == scalar.float32:
+        # The constants were obtained by looking at the output of
+        # python commands like:
+        #
+        # import numpy, theano
+        # dt='float32'  # or float64
+        # for i in xrange(750):
+        #     print i, repr(theano._asarray(1.0, dtype=dt) /
+        #                   (theano._asarray(1.0, dtype=dt) +
+        #                    numpy.exp(-theano._asarray([i,-i], dtype=dt))))
+
+        # float16 limits: -11.0, 7.0f
+        # We use the float32 limits for float16 for now as the
+        # computation will happend in float32 anyway.
+        if (node.inputs[0].type == scalar.float32 or
+                node.inputs[0].type == scalar.float16):
             return """%(z)s = %(x)s < -88.0f ? 0.0 : %(x)s > 15.0f ? 1.0f : 1.0f /(1.0f + exp(-%(x)s));""" % locals()
         elif node.inputs[0].type == scalar.float64:
             return """%(z)s = %(x)s < -709.0 ? 0.0 : %(x)s > 19.0 ? 1.0 : 1.0 /(1.0+exp(-%(x)s));""" % locals()
@@ -80,7 +96,7 @@ for i in xrange(750):
         x, = inp
         z, = out
         if (not theano.config.lib.amdlibm or
-            node.inputs[0].dtype != node.outputs[0].dtype):
+                node.inputs[0].dtype != node.outputs[0].dtype):
             raise theano.gof.utils.MethodNotDefined()
         dtype = node.inputs[0].dtype
         if dtype == 'float32' and self.amd_float32 is not None:
@@ -115,11 +131,11 @@ for i in xrange(750):
     @staticmethod
     def gen_graph():
         """
-        This method was used to generate the graph: sigmoid_prec.png in the doc
+        This method was used to generate the graph: sigmoid_prec.png in the doc.
+
         """
-        import matplotlib
         data = numpy.arange(-15, 15, .1)
-        val = 1/(1+numpy.exp(-data))
+        val = 1 / (1 + numpy.exp(-data))
 
         def hard_sigmoid(x):
             return theano.tensor.nnet.hard_sigmoid(x)
@@ -143,19 +159,19 @@ for i in xrange(750):
                              'doc', 'library', 'tensor', 'nnet',
                              'sigmoid_prec.png')
         plt.savefig(fname)
-        print "New picture saved at", fname
-        print val_ultra.max()
-        print val_ultra.min()
+        print("New picture saved at", fname)
+        print(val_ultra.max())
+        print(val_ultra.min())
 
 
 scalar_sigmoid = ScalarSigmoid(scalar.upgrade_to_float, name='scalar_sigmoid')
 sigmoid = elemwise.Elemwise(scalar_sigmoid, name='sigmoid')
 
 sigmoid_inplace = elemwise.Elemwise(
-        ScalarSigmoid(scalar.transfer_type(0)),
-        inplace_pattern={0: 0},
-        name='sigmoid_inplace',
-        )
+    ScalarSigmoid(scalar.transfer_type(0)),
+    inplace_pattern={0: 0},
+    name='sigmoid_inplace',
+)
 
 pprint.assign(sigmoid, printing.FunctionPrinter('sigmoid'))
 
@@ -163,6 +179,7 @@ pprint.assign(sigmoid, printing.FunctionPrinter('sigmoid'))
 class UltraFastScalarSigmoid(scalar.UnaryScalarOp):
     """
     This is just speed opt. Not for stability.
+
     """
     @staticmethod
     def st_impl(x):
@@ -228,14 +245,14 @@ pprint.assign(ultra_fast_sigmoid,
               printing.FunctionPrinter('ultra_fast_sigmoid'))
 
 
-#@opt.register_uncanonicalize
+# @opt.register_uncanonicalize
 @gof.local_optimizer([sigmoid])
 def local_ultra_fast_sigmoid(node):
     """
     When enabled, change all sigmoid to ultra_fast_sigmoid.
 
     For example do mode.including('local_ultra_fast_sigmoid')
-    or use the Theano flag optimizer_including=local_ultra_fast_sigmoid
+    or use the Theano flag optimizer_including=local_ultra_fast_sigmoid.
 
     This speeds up the sigmoid op by using an approximation.
 
@@ -246,60 +263,74 @@ def local_ultra_fast_sigmoid(node):
     if (isinstance(node.op, tensor.Elemwise) and
             node.op.scalar_op == scalar_sigmoid):
         out = ultra_fast_sigmoid(node.inputs[0])
+        copy_stack_trace(node.outputs[0], out)
 
         def values_eq_approx_remove_low_prec(a, b):
             # atol is found by trial/error.
             # Other test could fail without good reason.
             return tensor.TensorType.values_eq_approx(a, b, atol=0.02)
         # Let DebugMode know that there this opt approx the values.
-        out.values_eq_approx = values_eq_approx_remove_low_prec
+        out.tag.values_eq_approx = values_eq_approx_remove_low_prec
         return [out]
 theano.compile.optdb['uncanonicalize'].register("local_ultra_fast_sigmoid",
                                                 local_ultra_fast_sigmoid)
 
 
 def hard_sigmoid(x):
-    """An approximation of sigmoid.
+    """
+    An approximation of sigmoid.
 
     More approximate and faster than ultra_fast_sigmoid.
 
-    Approx in 3 parts: 0, scaled linear, 1
+    Approx in 3 parts: 0, scaled linear, 1.
 
     Removing the slope and shift does not make it faster.
 
     """
-    slope = 0.2
-    shift = 0.5
+    # Use the same dtype as determined by "upgrade_to_float",
+    # and perform computation in that dtype.
+    out_dtype = scalar.upgrade_to_float(scalar.Scalar(dtype=x.dtype))[0].dtype
+    slope = tensor.constant(0.2, dtype=out_dtype)
+    shift = tensor.constant(0.5, dtype=out_dtype)
     x = (x * slope) + shift
     x = tensor.clip(x, 0, 1)
     return x
 
 
-#@opt.register_uncanonicalize
+# @opt.register_uncanonicalize
 @gof.local_optimizer([sigmoid])
 def local_hard_sigmoid(node):
     if (isinstance(node.op, tensor.Elemwise) and
             node.op.scalar_op == scalar_sigmoid):
         out = hard_sigmoid(node.inputs[0])
+        copy_stack_trace(node.outputs[0], out)
 
         def values_eq_approx_remove_low_prec(a, b):
             # atol is found by trial/error.
             # Other test could fail without good reason.
             return tensor.TensorType.values_eq_approx(a, b, atol=0.1)
         # Let DebugMode know that there this opt approx the values.
-        out.values_eq_approx = values_eq_approx_remove_low_prec
+        out.tag.values_eq_approx = values_eq_approx_remove_low_prec
         return [out]
 theano.compile.optdb['uncanonicalize'].register("local_hard_sigmoid",
                                                 local_hard_sigmoid)
 
 
 class ScalarSoftplus(scalar.UnaryScalarOp):
+    """
+    This helps numerical stability.
+    """
     @staticmethod
     def static_impl(x):
         if x < -30.0:
             return 0.0
         if x > 30.0:
             return x
+        # If x is an int8 or uint8, numpy.exp will compute the result in
+        # half-precision (float16), where we want float32.
+        x_dtype = str(getattr(x, 'dtype', ''))
+        if x_dtype in ('int8', 'uint8'):
+            return numpy.log1p(numpy.exp(x, sig='f'))
         return numpy.log1p(numpy.exp(x))
 
     def impl(self, x):
@@ -313,11 +344,17 @@ class ScalarSoftplus(scalar.UnaryScalarOp):
     def c_code(self, node, name, inp, out, sub):
         x, = inp
         z, = out
-        if node.inputs[0].type == scalar.float32:
-            # These constants were obtained by looking at the output of python commands like:
-            #  for i in xrange(750):
-            #      print i, repr( numpy.log1p(numpy.exp(theano._asarray([i,-i], dtype=dt))))
-            # the boundary checks prevent us from generating inf
+        # These constants were obtained by looking at the output of
+        # python commands like:
+        #  for i in xrange(750):
+        #      print i, repr(numpy.log1p(numpy.exp(theano._asarray([i,-i], dtype=dt))))
+        # the boundary checks prevent us from generating inf
+
+        # float16 limits: -17.0, 6.0
+        # We use the float32 limits for float16 for now as the
+        # computation will happend in float32 anyway.
+        if (node.inputs[0].type == scalar.float32 or
+                node.inputs[0].type == scalar.float16):
             return """%(z)s = %(x)s < -103.0f ? 0.0 : %(x)s > 14.0f ? %(x)s : log1p(exp(%(x)s));""" % locals()
         elif node.inputs[0].type == scalar.float64:
             return """%(z)s = %(x)s < -745.0 ? 0.0 : %(x)s > 16.0 ? %(x)s : log1p(exp(%(x)s));""" % locals()
@@ -347,11 +384,18 @@ logsigm_to_softplus = gof.PatternSub(
     (tensor.log, (sigmoid, 'x')),
     (tensor.neg, (softplus, (tensor.neg, 'x'))),
     allow_multiple_clients=True,
+    values_eq_approx=values_eq_approx_remove_inf,
     skip_identities_fn=_skip_mul_1)
 
 
 def _is_1(expr):
-    """rtype bool. True iff expr is a constant close to 1
+    """
+
+    Returns
+    -------
+    bool
+        True iff expr is a constant close to 1.
+
     """
     try:
         v = opt.get_scalar_constant_value(expr)
@@ -366,12 +410,14 @@ log1msigm_to_softplus = gof.PatternSub(
             (sigmoid, 'x'))),
     (tensor.neg, (softplus, 'x')),
     allow_multiple_clients=True,
+    values_eq_approx=values_eq_approx_remove_inf,
     skip_identities_fn=_skip_mul_1)
 
 log1pexp_to_softplus = gof.PatternSub(
     (tensor.log1p,
      (tensor.exp, 'x')),
     (softplus, 'x'),
+    values_eq_approx=values_eq_approx_remove_inf,
     allow_multiple_clients=True)
 
 opt.register_stabilize(logsigm_to_softplus, name='logsigm_to_softplus')
@@ -381,12 +427,17 @@ opt.register_stabilize(log1pexp_to_softplus, name='log1pexp_to_softplus')
 
 def is_1pexp(t):
     """
-    If 't' is of the form (1+exp(x)), return (False, x).
-    Else return None.
+
+    Returns
+    -------
+    object
+        If 't' is of the form (1+exp(x)), return (False, x).
+        Else return None.
+
     """
     if t.owner and t.owner.op == tensor.add:
         scalars, scalar_inputs, nonconsts = \
-                opt.scalarconsts_rest(t.owner.inputs)
+            opt.scalarconsts_rest(t.owner.inputs)
         # scalar_inputs are potentially dimshuffled and fill'd scalars
         if len(nonconsts) == 1:
             maybe_exp = nonconsts[0]
@@ -413,22 +464,22 @@ def is_1pexp(t):
     return None
 
 
-AddConfigVar('warn.identify_1pexp_bug',
-        'Warn if Theano versions prior to 7987b51 (2011-12-18) could have '
-        'yielded a wrong result due to a bug in the is_1pexp function',
-        BoolParam(theano.configdefaults.warn_default('0.4.1')),
-        in_c_key=False)
-
-
 def is_exp(var):
     """
     Match a variable with either of the `exp(x)` or `-exp(x)` patterns.
 
-    :param var: The Variable to analyze.
+    Parameters
+    ----------
+    var
+        The Variable to analyze.
+
+    Returns
+    -------
+    tuple
+        A pair (b, x) with `b` a boolean set to True if `var` is of the
+        form `-exp(x)` and False if `var` is of the form `exp(x)`. If `var`
+        cannot be cast into either form, then return `None`.
 
-    :return: A pair (b, x) with `b` a boolean set to True if `var` is of the
-    form `-exp(x)` and False if `var` is of the form `exp(x)`. If `var` cannot
-    be cast into either form, then return `None`.
     """
     neg = False
     neg_info = is_neg(var)
@@ -443,10 +494,17 @@ def is_mul(var):
     """
     Match a variable with `x * y * z * ...`.
 
-    :param var: The Variable to analyze.
+    Parameters
+    ----------
+    var
+        The Variable to analyze.
+
+    Returns
+    -------
+    object
+        A list [x, y, z, ...] if `var` is of the form `x * y * z * ...`,
+        or None if `var` cannot be cast into this form.
 
-    :return: A list [x, y, z, ...] if `var` is of the form `x * y * z * ...`,
-    or None if `var` cannot be cast into this form.
     """
     if var.owner and var.owner.op == tensor.mul:
         return var.owner.inputs
@@ -479,9 +537,16 @@ def is_neg(var):
     """
     Match a variable with the `-x` pattern.
 
-    :param var: The Variable to analyze.
+    Parameters
+    ----------
+    var
+        The Variable to analyze.
+
+    Returns
+    -------
+    object
+        `x` if `var` is of the form `-x`, or None otherwise.
 
-    :return: `x` if `var` is of the form `-x`, or None otherwise.
     """
     apply = var.owner
     if not apply:
@@ -513,14 +578,16 @@ def is_neg(var):
 @opt.register_stabilize
 @gof.local_optimizer([tensor.true_div])
 def local_exp_over_1_plus_exp(node):
-    """exp(x)/(1+exp(x)) -> sigm(x)
+    """
+    exp(x)/(1+exp(x)) -> sigm(x)
     c/(1+exp(x)) -> c*sigm(-x)
+
     """
     # this optimization should be done for numerical stability
     # so we don't care to check client counts
     if node.op == tensor.true_div:
 
-        #find all the exp() terms in the numerator
+        # find all the exp() terms in the numerator
         num, denom = node.inputs
         num_exp_x, num_rest, num_neg = partition_num_or_denom(num, is_exp)
         denom_1pexp, denom_rest, \
@@ -560,20 +627,28 @@ def parse_mul_tree(root):
     """
     Parse a tree of multiplications starting at the given root.
 
-    :param root: The variable at the root of the tree.
+    Parameters
+    ----------
+    root
+        The variable at the root of the tree.
 
-    :return: A tree where each non-leaf node corresponds to a multiplication
-    in the computation of `root`, represented by the list of its inputs. Each
-    input is a pair [n, x] with `n` a boolean value indicating whether
-    sub-tree `x` should be negated.
+    Returns
+    -------
+    object
+        A tree where each non-leaf node corresponds to a multiplication
+        in the computation of `root`, represented by the list of its inputs.
+        Each input is a pair [n, x] with `n` a boolean value indicating whether
+        sub-tree `x` should be negated.
 
-    Examples:
+    Examples
+    --------
         x * y               -> [False, [[False, x], [False, y]]]
         -(x * y)            -> [True, [[False, x], [False, y]]]
         -x * y              -> [False, [[True, x], [False, y]]]
         -x                  -> [True, x]
         (x * y) * -z        -> [False, [[False, [[False, x], [False, y]]],
                                         [True, z]]]
+
     """
     # Is it a multiplication?
     mul_info = is_mul(root)
@@ -589,34 +664,41 @@ def parse_mul_tree(root):
             return [not neg, sub_tree]
     else:
         # Recurse into inputs.
-        return [False, map(parse_mul_tree, mul_info)]
+        return [False, list(map(parse_mul_tree, mul_info))]
 
 
 def replace_leaf(arg, leaves, new_leaves, op, neg):
     """
-    Attempts to replace a leaf of a multiplication tree.
+    Attempt to replace a leaf of a multiplication tree.
 
     We search for a leaf in `leaves` whose argument is `arg`, and if we find
     one, we remove it from `leaves` and add to `new_leaves` a leaf with
     argument `arg` and variable `op(arg)`.
 
-    :param arg: The argument of the leaf we are looking for.
+    Parameters
+    ----------
+    arg
+        The argument of the leaf we are looking for.
+    leaves
+        List of leaves to look into. Each leaf should be a pair
+        (x, l) with `x` the argument of the Op found in the leaf, and `l` the
+        actual leaf as found in a multiplication tree output by `parse_mul_tree`
+        (i.e. a pair [boolean, variable]).
+    new_leaves
+        If a replacement occurred, then the leaf is removed from `leaves`
+        and added to the list `new_leaves` (after being modified by `op`).
+    op
+        A function that, when applied to `arg`, returns the Variable
+        we want to replace the original leaf variable with.
+    neg : bool
+        If True, then the boolean value associated to the leaf should
+        be swapped. If False, then this value should remain unchanged.
 
-    :param leaves: List of leaves to look into. Each leaf should be a pair
-    (x, l) with `x` the argument of the Op found in the leaf, and `l` the
-    actual leaf as found in a multiplication tree output by `parse_mul_tree`
-    (i.e. a pair [boolean, variable]).
+    Returns
+    -------
+    bool
+        True if a replacement occurred, or False otherwise.
 
-    :param new_leaves: If a replacement occurred, then the leaf is removed from
-    `leaves` and added to the list `new_leaves` (after being modified by `op`).
-
-    :param op: A function that, when applied to `arg`, returns the Variable
-    we want to replace the original leaf variable with.
-
-    :param neg: If True, then the boolean value associated to the leaf should
-    be swapped. If False, then this value should remain unchanged.
-
-    :return: True if a replacement occurred, or False otherwise.
     """
     for idx, x in enumerate(leaves):
         if x[0] == arg:
@@ -632,12 +714,19 @@ def simplify_mul(tree):
     """
     Simplify a multiplication tree.
 
-    :param tree: A multiplication tree (as output by `parse_mul_tree`).
+    Parameters
+    ----------
+    tree
+        A multiplication tree (as output by `parse_mul_tree`).
+
+    Returns
+    -------
+    object
+        A multiplication tree computing the same output as `tree` but without
+        useless multiplications by 1 nor -1 (identified by leaves of the form
+        [False, None] or [True, None] respectively). Useless multiplications
+        (with less than two inputs) are also removed from the tree.
 
-    :return: A multiplication tree computing the same output as `tree` but
-    without useless multiplications by 1 nor -1 (identified by leaves of the
-    form [False, None] or [True, None] respectively). Useless multiplications
-    (with less than two inputs) are also removed from the tree.
     """
     neg, inputs = tree
     if isinstance(inputs, list):
@@ -660,7 +749,7 @@ def simplify_mul(tree):
             rval = [neg, s_inputs]
     else:
         rval = tree
-    #print 'simplify_mul: %s -> %s' % (tree, rval)
+    # print 'simplify_mul: %s -> %s' % (tree, rval)
     return rval
 
 
@@ -669,12 +758,18 @@ def compute_mul(tree):
     Compute the Variable that is the output of a multiplication tree.
 
     This is the inverse of the operation performed by `parse_mul_tree`, i.e.
-        compute_mul(parse_mul_tree(tree)) == tree
+    compute_mul(parse_mul_tree(tree)) == tree.
 
-    :param tree: A multiplication tree (as output by `parse_mul_tree`).
+    Parameters
+    ----------
+    tree
+        A multiplication tree (as output by `parse_mul_tree`).
+
+    Returns
+    -------
+    object
+        A Variable that computes the multiplication represented by the tree.
 
-    :return: A Variable that computes the multiplication represented by the
-    tree.
     """
     neg, inputs = tree
     if inputs is None:
@@ -683,7 +778,7 @@ def compute_mul(tree):
             'call `simplify_mul` on the tree first?')
     elif isinstance(inputs, list):
         # Recurse through inputs.
-        rval = tensor.mul(*map(compute_mul, inputs))
+        rval = tensor.mul(*list(map(compute_mul, inputs)))
     else:
         rval = inputs
     if neg:
@@ -702,32 +797,38 @@ def perform_sigm_times_exp(tree, exp_x=None, exp_minus_x=None, sigm_x=None,
     by replacing matching pairs (exp, sigmoid) with the desired optimized
     version.
 
-    :param tree: The sub-tree to operate on.
+    Parameters
+    ----------
+    tree
+        The sub-tree to operate on.
+    exp_x
+        List of arguments x so that `exp(x)` exists somewhere in the whole
+        multiplication tree. Each argument is a pair (x, leaf) with `x` the
+        argument of the exponential, and `leaf` the corresponding leaf in the
+        multiplication tree (of the form [n, exp(x)] -- see `parse_mul_tree`).
+        If None, this argument is initialized to an empty list.
+    exp_minus_x
+        Similar to `exp_x`, but for `exp(-x)`.
+    sigm_x
+        Similar to `exp_x`, but for `sigmoid(x)`.
+    sigm_minus_x
+        Similar to `exp_x`, but for `sigmoid(-x)`.
+    parent
+        Parent of `tree` (None if `tree` is the global root).
+    child_idx
+        Index of `tree` in its parent's inputs (None if `tree` is the global
+        root).
+    full_tree
+        The global multiplication tree (should not be set except by recursive
+        calls to this function). Used for debugging only.
 
-    :exp_x: List of arguments x so that `exp(x)` exists somewhere in the whole
-    multiplication tree. Each argument is a pair (x, leaf) with `x` the
-    argument of the exponential, and `leaf` the corresponding leaf in the
-    multiplication tree (of the form [n, exp(x)] -- see `parse_mul_tree`).
-    If None, this argument is initialized to an empty list.
+    Returns
+    -------
+    bool
+        True if a modification was performed somewhere in the whole multiplication
+        tree, or False otherwise.
 
-    :param exp_minus_x: Similar to `exp_x`, but for `exp(-x)`.
-
-    :param sigm_x: Similar to `exp_x`, but for `sigmoid(x)`.
-
-    :param sigm_minus_x: Similar to `exp_x`, but for `sigmoid(-x)`.
-
-    :param parent: Parent of `tree` (None if `tree` is the global root).
-
-    :param child_idx: Index of `tree` in its parent's inputs (None if `tree` is
-    the global root).
-
-    :param full_tree: The global multiplication tree (should not be set except
-    by recursive calls to this function). Used for debugging only.
-
-    :return: True if a modification was performed somewhere in the whole
-    multiplication tree, or False otherwise.
     """
-
     if exp_x is None:
         exp_x = []
     if exp_minus_x is None:
@@ -739,22 +840,22 @@ def perform_sigm_times_exp(tree, exp_x=None, exp_minus_x=None, sigm_x=None,
     if full_tree is None:
         full_tree = tree
     if False:  # Debug code.
-        print '<perform_sigm_times_exp>'
-        print '  full_tree   = %s' % full_tree
-        print '  tree        = %s' % tree
-        print '  exp_x       = %s' % exp_x
-        print '  exp_minus_x = %s' % exp_minus_x
-        print '  sigm_x      = %s' % sigm_x
-        print '  sigm_minus_x= %s' % sigm_minus_x
+        print('<perform_sigm_times_exp>')
+        print('  full_tree   = %s' % full_tree)
+        print('  tree        = %s' % tree)
+        print('  exp_x       = %s' % exp_x)
+        print('  exp_minus_x = %s' % exp_minus_x)
+        print('  sigm_x      = %s' % sigm_x)
+        print('  sigm_minus_x= %s' % sigm_minus_x)
     neg, inputs = tree
     if isinstance(inputs, list):
         # Recurse through inputs of the multiplication.
         rval = False
         for sub_idx, sub_tree in enumerate(inputs):
             rval |= perform_sigm_times_exp(
-                    tree=sub_tree, parent=tree, child_idx=sub_idx,
-                    exp_x=exp_x, exp_minus_x=exp_minus_x, sigm_x=sigm_x,
-                    sigm_minus_x=sigm_minus_x, full_tree=full_tree)
+                tree=sub_tree, parent=tree, child_idx=sub_idx,
+                exp_x=exp_x, exp_minus_x=exp_minus_x, sigm_x=sigm_x,
+                sigm_minus_x=sigm_minus_x, full_tree=full_tree)
         return rval
     else:
         # Reached a leaf: if it is an exponential or a sigmoid, then we
@@ -811,6 +912,7 @@ def local_sigm_times_exp(node):
     """
     exp(x) * sigm(-x) -> sigm(x)
     exp(-x) * sigm(x) -> sigm(-x)
+
     """
     # Bail early if it is not a multiplication.
     if node.op != tensor.mul:
@@ -826,7 +928,10 @@ def local_sigm_times_exp(node):
     # get rid of them.
     mul_tree = simplify_mul(mul_tree)
     # Recompute final output based on the updated tree.
-    return [compute_mul(mul_tree)]
+    out = compute_mul(mul_tree)
+    # keep the stack trace
+    copy_stack_trace(node.outputs[0], out)
+    return [out]
 
 
 @opt.register_stabilize
@@ -834,6 +939,7 @@ def local_sigm_times_exp(node):
 def local_inv_1_plus_exp(node):
     """
     1/(1+exp(x)) -> sigm(-x)
+
     """
     # this optimization should be done for numerical stability
     # so we don't care to check client counts
@@ -841,15 +947,22 @@ def local_inv_1_plus_exp(node):
         inv_arg = node.inputs[0]
         if inv_arg.owner and inv_arg.owner.op == tensor.add:
             scalars, scalar_inputs, nonconsts = \
-                    opt.scalarconsts_rest(inv_arg.owner.inputs)
+                opt.scalarconsts_rest(inv_arg.owner.inputs)
             # scalar_inputs are potentially dimshuffled and fill'd scalars
             if len(nonconsts) == 1:
                 if nonconsts[0].owner and nonconsts[0].owner.op == tensor.exp:
                     if scalars and numpy.allclose(numpy.sum(scalars), 1):
-                        return opt._fill_chain(
-                                sigmoid(
-                                    tensor.neg(nonconsts[0].owner.inputs[0])),
-                                scalar_inputs)
+                        out = opt._fill_chain(
+                            sigmoid(
+                                tensor.neg(nonconsts[0].owner.inputs[0])),
+                            scalar_inputs)
+                        # keep combined stack traces of
+                        #     exp(x):           nonconsts[0],
+                        #     1 + exp(x):       inv_arg,
+                        #     1 / (1 + exp(x)): node.outputs[0]
+                        copy_stack_trace(
+                            [nonconsts[0], inv_arg, node.outputs[0]], out)
+                        return out
 
 # Registration is below, and conditional.
 
@@ -858,6 +971,7 @@ def local_inv_1_plus_exp(node):
 def local_1msigmoid(node):
     """
     1-sigm(x) -> sigm(-x)
+
     """
     if node.op == tensor.sub:
         sub_l, sub_r = node.inputs
@@ -866,10 +980,12 @@ def local_1msigmoid(node):
         if sub_r.owner and sub_r.owner.op == sigmoid:
             try:
                 val_l = opt.get_scalar_constant_value(sub_l)
-            except Exception, e:
+            except Exception:
                 return
             if numpy.allclose(numpy.sum(val_l), 1):
-                return [sigmoid(-sub_r.owner.inputs[0])]
+                out = sigmoid(-sub_r.owner.inputs[0])
+                copy_stack_trace([sub_r, node.outputs[0]], out)
+                return [out]
 
 register_local_1msigmoid = False
 # This is False because the Stabilize pattern above
@@ -891,11 +1007,10 @@ if 0:
     @opt.register_stabilize
     @gof.local_optimizer([tensor.mul])
     def local_sigm_gest(node):
-        print "CANONICALIZE"
-        print sigm_canonicalize(node)
+        print("CANONICALIZE")
+        print(sigm_canonicalize(node))
 
     def sigm_canonicalize(node):
-        add = tensor.add
         mul = tensor.mul
         div = tensor.true_div
 
