@@ -1,4 +1,6 @@
+from __future__ import print_function
 from copy import copy
+from itertools import product as itertools_product
 from unittest import TestCase
 
 import numpy
@@ -6,12 +8,12 @@ from numpy import (arange, array, common_type, complex64, complex128, float32,
                   float64, newaxis, shape, transpose, zeros)
 from numpy.testing import assert_array_almost_equal
 
+from six.moves import xrange
+
 import theano
 import theano.tensor as T
-from theano import tensor, Param, shared, config
+from theano import tensor, In, shared, config
 from theano.compat import exc_message
-from theano.gof.python25 import product as itertools_product
-from theano.gof.python25 import any
 from theano.printing import pp
 from theano.tensor.blas import (_dot22, _dot22scalar, res_is_a, _as_scalar,
                                 _is_real_matrix, _gemm_canonicalize,
@@ -19,10 +21,10 @@ from theano.tensor.blas import (_dot22, _dot22scalar, res_is_a, _as_scalar,
                                 gemm_inplace, gemm_no_inplace,
                                 InconsistencyError, Ger, ger, ger_destructive)
 from theano.tests import unittest_tools
-from test_basic import (_approx_eq, as_tensor_variable, inplace_func,
-        compile, inplace)
-        #, constant, eval_outputs)
+from .test_basic import (as_tensor_variable, inplace_func,
+                        compile, inplace)
 import theano.tensor.blas_scipy
+from theano.tests.unittest_tools import attr
 
 
 if config.mode == 'FAST_COMPILE':
@@ -50,7 +52,6 @@ class t_gemm(TestCase):
     """
     def setUp(self):
         unittest_tools.seed_rng()
-        _approx_eq.debug = 0
         Gemm.debug = False
 
     @staticmethod
@@ -72,10 +73,10 @@ class t_gemm(TestCase):
             b = numpy.asarray(b_, dtype=dtype)
 
             def cmp_linker(z, a, x, y, b, l):
-                z, a, x, y, b = [numpy.asarray(p) for p in z, a, x, y, b]
+                z, a, x, y, b = [numpy.asarray(p) for p in (z, a, x, y, b)]
                 z_orig = z.copy()
                 tz, ta, tx, ty, tb = [as_tensor_variable(p).type()
-                                      for p in z, a, x, y, b]
+                                      for p in (z, a, x, y, b)]
 
                 f = inplace_func([tz, ta, tx, ty, tb],
                                  gemm_inplace(tz, ta, tx, ty, tb),
@@ -83,9 +84,8 @@ class t_gemm(TestCase):
                 new_z = f(z, a, x, y, b)
                 z_after = self._gemm(z_orig, a, x, y, b)
 
-                #print z_orig, z_after, z, type(z_orig), type(z_after), type(z)
-                #_approx_eq.debug = 1
-                self.assertTrue(_approx_eq(z_after, z))
+                # print z_orig, z_after, z, type(z_orig), type(z_after), type(z)
+                unittest_tools.assert_allclose(z_after, z)
                 if a == 0.0 and b == 1.0:
                     return
                 elif z_orig.size == 0:
@@ -105,7 +105,7 @@ class t_gemm(TestCase):
         Gemm.debug = True
         try:
             g = gemm_inplace([1.], 1., [1.], [1.], 1.)
-        except TypeError, e:
+        except TypeError as e:
             if exc_message(e) is Gemm.E_rank:
                 return
         self.fail()
@@ -113,7 +113,7 @@ class t_gemm(TestCase):
     def test0(self):
         try:
             self.cmp(1., 0., 1.0, 1.0, 1.0)
-        except TypeError, e:
+        except TypeError as e:
             if exc_message(e) is Gemm.E_rank:
                 return
         self.fail()
@@ -121,7 +121,7 @@ class t_gemm(TestCase):
     def test2(self):
         try:
             self.cmp(2., 1.0, [3, 2, 1.], [[1], [2], [3.]], 1.0)
-        except TypeError, e:
+        except TypeError as e:
             self.assertTrue(exc_message(e) == Gemm.E_rank)
             return
         self.fail()
@@ -150,7 +150,6 @@ class t_gemm(TestCase):
                  self.rand(3, 5), self.rand(5, 4), -1.0)
 
     def test10(self):
-        _approx_eq.debug = 1
         self.cmp(self.rand(3, 4), -1.0, self.rand(3, 5), self.rand(5, 4), 0.0)
 
     def test11(self):
@@ -178,7 +177,7 @@ class t_gemm(TestCase):
         lr2 = T.constant(2).astype(config.floatX)
         l2_reg = T.constant(0.0001).astype(config.floatX)
 
-        #test constant merge with gemm
+        # test constant merge with gemm
         f = theano.function([a, b], updates=[(s, lr1 * T.dot(a, b) +
                                                 l2_reg * lr2 * s)],
                             mode=mode_not_fast_compile).maker.fgraph.toposort()
@@ -188,7 +187,7 @@ class t_gemm(TestCase):
         assert len(f) == 1
         assert f[0].op == gemm_inplace
 
-        #test factored scalar with merge
+        # test factored scalar with merge
         f = theano.function([a, b], updates=[(s, lr1 * (T.dot(a, b) -
                                                         l2_reg * s))],
                             mode=mode_not_fast_compile).maker.fgraph.toposort()
@@ -198,7 +197,7 @@ class t_gemm(TestCase):
         assert len(f) == 1
         assert f[0].op == gemm_inplace
 
-        #test factored scalar with merge and neg
+        # test factored scalar with merge and neg
         f = theano.function([a, b],
                             updates=[(s, s - lr1 * (s * .0002 + T.dot(a, b)))],
                             mode=mode_not_fast_compile).maker.fgraph.toposort()
@@ -213,7 +212,7 @@ class t_gemm(TestCase):
         Z = as_tensor_variable(self.rand(2, 2))
         try:
             gemm_inplace(Z, 1.0, Z, Z, 1.0)
-        except InconsistencyError, e:
+        except InconsistencyError as e:
             if exc_message(e) == Gemm.E_z_uniq:
                 return
         self.fail()
@@ -224,7 +223,7 @@ class t_gemm(TestCase):
         A = as_tensor_variable(self.rand(2, 2))
         try:
             gemm_inplace(Z, 1.0, A, inplace.transpose_inplace(Z), 1.0)
-        except InconsistencyError, e:
+        except InconsistencyError as e:
             if exc_message(e) == Gemm.E_z_uniq:
                 return
         self.fail()
@@ -235,7 +234,7 @@ class t_gemm(TestCase):
         A = as_tensor_variable(self.rand(2, 2))
         try:
             gemm_inplace(Z, 1.0, inplace.transpose_inplace(Z), A, 1.0)
-        except InconsistencyError, e:
+        except InconsistencyError as e:
             if exc_message(e) == Gemm.E_z_uniq:
                 return
         self.fail()
@@ -246,7 +245,7 @@ class t_gemm(TestCase):
         A = as_tensor_variable(self.rand(2, 2))
         try:
             gemm_inplace(Z, 1.0, Z, A, 1.0)
-        except InconsistencyError, e:
+        except InconsistencyError as e:
             if exc_message(e) == Gemm.E_z_uniq:
                 return
         self.fail()
@@ -269,28 +268,25 @@ class t_gemm(TestCase):
 
         def t(z, x, y, a=1.0, b=0.0, l='c|py', dt='float64'):
             z, a, x, y, b = [theano._asarray(p, dtype=dt)
-                             for p in z, a, x, y, b]
+                             for p in (z, a, x, y, b)]
             z_orig = z.copy()
             z_after = self._gemm(z, a, x, y, b)
 
-            tz, ta, tx, ty, tb = [shared(p) for p in z, a, x, y, b]
+            tz, ta, tx, ty, tb = [shared(p) for p in (z, a, x, y, b)]
 
-            #f = inplace_func([tz,ta,tx,ty,tb], gemm_inplace(tz,ta,tx,ty,tb),
+            # f = inplace_func([tz,ta,tx,ty,tb], gemm_inplace(tz,ta,tx,ty,tb),
             #                 mode = compile.Mode(optimizer = None, linker=l))
             #f(z, a, x, y, b)
             f = inplace_func([], gemm_inplace(tz, ta, tx, ty, tb),
                              mode=compile.Mode(optimizer=None, linker=l))
             f()
-            self.assertTrue(_approx_eq(z_after, tz.get_value(borrow=True)),
-                            (z_orig, z_after, z, z_after - z))
+            unittest_tools.assert_allclose(z_after, tz.get_value(borrow=True))
             f()
-            self.assertTrue(_approx_eq(z_after, tz.get_value(borrow=True)),
-                            (z_orig, z_after, z, z_after - z))
+            unittest_tools.assert_allclose(z_after, tz.get_value(borrow=True))
             f()
-            self.assertTrue(_approx_eq(z_after, tz.get_value(borrow=True)),
-                            (z_orig, z_after, z, z_after - z))
+            unittest_tools.assert_allclose(z_after, tz.get_value(borrow=True))
 
-            #tz.value *= 0 # clear z's value
+            # tz.value *= 0 # clear z's value
             y_T = ty.get_value(borrow=True).T
             ty.set_value(tx.get_value(borrow=True).T, borrow=True)
             tx.set_value(y_T, borrow=True)
@@ -298,7 +294,7 @@ class t_gemm(TestCase):
             f()
             # test that the transposed version of multiplication gives
             # same answer
-            self.assertTrue(_approx_eq(z_after, tz.get_value(borrow=True).T))
+            unittest_tools.assert_allclose(z_after, tz.get_value(borrow=True).T)
 
         t(C, A, B)
         t(C.T, A, B)
@@ -318,7 +314,7 @@ class t_gemm(TestCase):
 
         try:
             t(C.T, A[:2, :], B[:, :2].T)
-        except ValueError, e:
+        except ValueError as e:
             if exc_message(e).find('aligned') >= 0:
                 return
         self.fail()
@@ -332,14 +328,14 @@ class t_gemm(TestCase):
 
         def t(z, x, y, a=1.0, b=0.0, l='c|py', dt='float64'):
             z, a, x, y, b = [theano._asarray(p, dtype=dt)
-                             for p in z, a, x, y, b]
+                             for p in (z, a, x, y, b)]
             z_orig = z.copy()
             z_after = numpy.zeros_like(z_orig)
             for i in xrange(3):
                 z_after[:, :, i] = self._gemm(z[:, :, i], a,
                                               x[:, :, i], y[:, :, i], b)
 
-            tz, ta, tx, ty, tb = [shared(p) for p in z, a, x, y, b]
+            tz, ta, tx, ty, tb = [shared(p) for p in (z, a, x, y, b)]
             for i in xrange(3):
                 f_i = inplace_func([],
                         gemm_inplace(tz[:, :, i],
@@ -384,7 +380,7 @@ def test_res_is_a():
     assert not res_is_a(a + a, T.sqrt)
     assert res_is_a(T.sqrt(a + a), T.sqrt)
 
-    #leave the maxclients  stuff untested because it requires being in an fgraph.
+    # leave the maxclients  stuff untested because it requires being in an fgraph.
 
 
 class t_as_scalar(TestCase):
@@ -438,7 +434,7 @@ class T_real_matrix(TestCase):
 
 
 def fail(msg):
-    print 'FAIL', msg
+    print('FAIL', msg)
     assert False
 
 
@@ -461,7 +457,7 @@ def just_gemm(i, o, ishapes=[(4, 3), (3, 5), (4, 5), (), ()],
               max_graphlen=0, expected_nb_gemm=1):
     try:
         f = inplace_func(
-                [Param(ii, mutable=True, allow_downcast=True) for ii in i],
+                [In(ii, mutable=True, allow_downcast=True) for ii in i],
                 o,
                 mode='FAST_RUN',
                 on_unused_input='ignore')
@@ -500,7 +496,7 @@ def just_gemm(i, o, ishapes=[(4, 3), (3, 5), (4, 5), (), ()],
                           max_abs_err)
     except Failure:
         for node in f.maker.fgraph.toposort():
-            print 'GRAPH', node
+            print('GRAPH', node)
         raise
 
 
@@ -515,13 +511,13 @@ def test_gemm_opt0():
     just_gemm([X, Y, Z, a, b], [a * T.dot(X, Y) - b * Z])
     just_gemm([X, Y, Z, a, b], [b * Z - a * T.dot(X, Y)])
 
-    #with transposes (transposes should be pushed through dot in canonicalize)
+    # with transposes (transposes should be pushed through dot in canonicalize)
     just_gemm([X, Y, Z, a, b], [b * Z.T - a * T.dot(Y.T, X.T)])
     just_gemm([X, Y, Z, a, b], [b * Z.T + a * b * T.dot(X, Y).T])
     just_gemm([X, Y, Z, a, b], [b * Z + a * T.dot(X, Y).T],
             ishapes=[(5, 3), (3, 4), (4, 5), (), ()])
 
-    #with N multiplications instead of just one
+    # with N multiplications instead of just one
     just_gemm([X, Y, Z, a, b], [(b * b) * Z * a + (a * a) * T.dot(X, Y) * b])
     just_gemm([X, Y, Z, a, b], [Z + T.dot(X, Y)])
     just_gemm([X, Y, Z, a, b], [Z * b + T.dot(X, Y)])
@@ -547,7 +543,7 @@ def test_gemm_opt_double_gemm():
     o = [(a * T.dot(X, Y)
         + gemm_inplace(Z, b, S.T, R.T, T.constant(1.0).astype(config.floatX)))]
     try:
-        f = inplace_func([Param(ii, mutable=True) for ii in i], o,
+        f = inplace_func([In(ii, mutable=True) for ii in i], o,
                 mode='FAST_RUN', on_unused_input='ignore')
         for node in f.maker.fgraph.apply_nodes:
             if isinstance(node.op, T.Dot):
@@ -556,7 +552,7 @@ def test_gemm_opt_double_gemm():
                 raise Failure('_dot22 in graph')
         g = inplace_func(i, o, mode=compile.Mode(linker='py', optimizer=None),
                 on_unused_input='ignore')
-        #for node in g.maker.fgraph.apply_nodes:
+        # for node in g.maker.fgraph.apply_nodes:
         #    if node.op == gemm_inplace: raise Failure('gemm_inplace in graph')
 
         rng = numpy.random.RandomState(unittest_tools.fetch_seed(234))
@@ -575,7 +571,7 @@ def test_gemm_opt_double_gemm():
                 max_abs_err)
     except Failure:
         for node in f.maker.fgraph.toposort():
-            print 'GRAPH', node
+            print('GRAPH', node)
         raise
 
 
@@ -622,7 +618,7 @@ def test_gemm_canonicalize():
 
     can = []
     _gemm_canonicalize((-d) * X - (a * X + Y - b * Z * c), 1.0, can, 0)
-    #print can
+    # print can
     assert can[0][0].owner.op == T.neg
     assert can[0][0].owner.inputs[0] == d
     assert can[0][1] == X
@@ -687,13 +683,13 @@ def test_gemm_nested():
             ishapes=[(2, 3), (3, 4), (2, 4), (2, 3), (3, 4), (
                 2, 4), (), (), (), ()],
             max_graphlen=1)
-    #print "---------------------"
+    # print "---------------------"
     just_gemm([X, Y, Z, R, S, U, a, b, c, d],
             [a * Z - b * (c * T.dot(X, Y) + d * Z + c * Z)],
             ishapes=[(2, 3), (3, 4), (2, 4), (2, 3), (3, 4), (
                 2, 4), (), (), (), ()],
             max_graphlen=1)
-    #print "---------------------"
+    # print "---------------------"
     just_gemm([X, Y, Z, R, S, U, a, b, c, d],
             [a * Z - b * (c * T.dot(X, Y) + d * Z + c * U)],
             ishapes=[(2, 3), (3, 4), (2, 4), (2, 3), (3, 4), (
@@ -704,7 +700,7 @@ def test_gemm_nested():
 def test_gemm_opt_wishlist():
     X, Y, Z, a, b = T.matrix(), T.matrix(), T.matrix(), T.scalar(), T.scalar()
 
-    #with >2 additions of the same T.dot(X,Y term
+    # with >2 additions of the same T.dot(X,Y term
     just_gemm([X, Y, Z, a, b],
               [(b * b) * Z * a + (a * a) * T.dot(X, Y) + b * T.dot(X, Y)])
 
@@ -732,7 +728,7 @@ def test_gemm_with_vector():
     my_just_gemm([v + a * T.dot(X, Y) - b * Z])
     my_just_gemm([v + b * Z - a * T.dot(X, Y)])
 
-    #with N multiplications instead of just one
+    # with N multiplications instead of just one
     my_just_gemm([v + (b * b) * Z * a + (a * a) * T.dot(X, Y) * b])
     my_just_gemm([v + Z + T.dot(X, Y)])
     my_just_gemm([v + Z * b + T.dot(X, Y)])
@@ -803,8 +799,8 @@ def test_gemm_unrolled():
 
 
 def test_inplace0():
-    #should fail to insert gemm_inplace because gemm_inplace would
-    #create cycles
+    # should fail to insert gemm_inplace because gemm_inplace would
+    # create cycles
     X, Y, Z, a, b = T.matrix('X'), T.matrix('Y'), T.matrix('Z'), T.scalar(
         'a'), T.scalar('b')
     R, S, c = T.matrix('R'), T.matrix('S'), T.scalar('c')
@@ -812,7 +808,7 @@ def test_inplace0():
     f = inplace_func([Z, b, R, S],
             [Z * (Z + b * T.dot(R, S).T)], mode='FAST_RUN')
     if (gemm_inplace in [n.op for n in f.maker.fgraph.apply_nodes]):
-        print pp(f.maker.fgraph.outputs[0])
+        print(pp(f.maker.fgraph.outputs[0]))
         raise Failure('gemm_inplace in graph')
     assert gemm_no_inplace in [n.op for n in f.maker.fgraph.apply_nodes]
 
@@ -830,7 +826,7 @@ def test_inplace1():
     # with > 2 terms in the overall addition
     f = inplace_func([X, Y, Z],
             [Z + Z + T.dot(X, Y)], mode='FAST_RUN')
-    #theano.printing.debugprint(f)
+    # theano.printing.debugprint(f)
     # it doesn't work inplace because we didn't mark Z as mutable input
     assert [n.op for n in f.maker.fgraph.apply_nodes] == [gemm_no_inplace]
 
@@ -862,11 +858,12 @@ def test_dot22():
             cmp((0, 0), (0, 0))
 
 
+@attr('slow')
 def test_dot22scalar():
-    ## including does not seem to work for 'local_dot_to_dot22' and
-    ## 'local_dot22_to_dot22scalar'
-    ## TODO: exclude other optimizations in BlasOpt?
-    #m = theano.compile.get_default_mode().including('local_dot_to_dot22',
+    # including does not seem to work for 'local_dot_to_dot22' and
+    # 'local_dot22_to_dot22scalar'
+    # TODO: exclude other optimizations in BlasOpt?
+    # m = theano.compile.get_default_mode().including('local_dot_to_dot22',
     #                           'local_dot22_to_dot22scalar','specialize')
     #m = theano.compile.get_default_mode().including('BlasOpt', 'specialize')
     rng = numpy.random.RandomState(unittest_tools.fetch_seed())
@@ -883,8 +880,10 @@ def test_dot22scalar():
                     def check_dot22scalar(func, len_topo_scalar=-1):
                         topo = func.maker.fgraph.toposort()
                         ops = [x.op for x in topo]
+                        classes = [type(x.op) for x in topo]
                         dtype4_upcast = theano.scalar.upcast(dtype4, dtype1,
                                                              dtype2)
+
                         if dtype1 == dtype2 == dtype3 == dtype4_upcast:
                             if len_topo_scalar > 0:
                                 assert len(topo) == len_topo_scalar
@@ -941,8 +940,8 @@ def test_dot22scalar():
                         check_dot22scalar(f, 2)
                         f(av, bv, cv)
 
-                        ## Here, canonicalize also seems needed
-                        ## TODO: add only the optimizations needed?
+                        # Here, canonicalize also seems needed
+                        # TODO: add only the optimizations needed?
                         m2 = mode_blas_opt.including('canonicalize')
                         f = theano.function([a, b, c],
                                             cst2 * c * cst * T.dot(a, b),
@@ -963,7 +962,7 @@ def test_dot22scalar():
                                                 cst * c * a * T.dot(a, b),
                                                 mode=mode_blas_opt)
                             topo = f.maker.fgraph.toposort()
-                            #currently the canonizer don't always
+                            # currently the canonizer don't always
                             # merge all Mul together...  dot22scalar
                             # optimizer does not do a recursive search
                             # therefore, it doesn't find potential
@@ -1027,7 +1026,7 @@ def test_local_dot22_to_dot22scalar():
     m = T.dmatrix()
     r = T.drow()
     for idx, node in enumerate([
-        #Old working cases
+        # Old working cases
         T.mul(_dot22(A, A), x),
         T.mul(_dot22(A, A), x, y),
         T.mul(_dot22(A, A), x, r),
@@ -1053,7 +1052,7 @@ def test_local_dot22_to_dot22scalar():
         T.mul(_dot22(A, A), T.mul(m, y, z), m),
         T.mul(_dot22(A, A), m, T.mul(m, y, z)),
 
-        #Case that opt later in gh-1515
+        # Case that opt later in gh-1515
         T.mul(_dot22(A, A), (r * m), (m * x)),
     ]):
         node2 = theano.tensor.blas.local_dot22_to_dot22scalar.transform(
@@ -1082,7 +1081,7 @@ def test_dot_w_self():
 
 
 ###############################################################################
-## Tests for Gemv
+# Tests for Gemv
 ###############################################################################
 
 class TestGemv(TestCase, unittest_tools.TestOptimizationMixin):
@@ -1095,7 +1094,7 @@ class TestGemv(TestCase, unittest_tools.TestOptimizationMixin):
 
         # Assert that the dot was optimized somehow
         self.assertFunctionContains0(f, T.dot)
-        self.assertFunctionContains1(f, Gemv(False))
+        self.assertFunctionContains1(f, Gemv(True))
 
         # Assert they produce the same output
         assert numpy.allclose(f(), numpy.dot(v.get_value(), w.get_value()))
@@ -1161,7 +1160,7 @@ class TestGemv(TestCase, unittest_tools.TestOptimizationMixin):
         assert isinstance(topo[0].op, Gemv)
         assert topo[0].op.inplace == False
 
-        #test the inplace version
+        # test the inplace version
         g = theano.function([], [], updates=[(v2, v2 + theano.dot(m, v1))],
                             mode=mode_blas_opt)
 
@@ -1186,6 +1185,7 @@ class TestGemv(TestCase, unittest_tools.TestOptimizationMixin):
         assert numpy.allclose(v2.get_value(),
                 numpy.dot(m.get_value(), v1.get_value()) + v2_orig)
 
+    @attr('slow')
     def test_gemv1(self):
         self.t_gemv1((3, 2))
         self.t_gemv1((0, 2))
@@ -1211,7 +1211,7 @@ class TestGemv(TestCase, unittest_tools.TestOptimizationMixin):
         assert sum(isinstance(node.op, Gemv) for node in topo) == 1
         assert topo[-1].op.inplace == False
 
-        #test the inplace version
+        # test the inplace version
         g = theano.function([], [], updates=[(v2, v2 + theano.dot(v1, m))],
                             mode=mode_blas_opt)
 
@@ -1235,6 +1235,35 @@ class TestGemv(TestCase, unittest_tools.TestOptimizationMixin):
         assert numpy.allclose(v2.get_value(),
                 numpy.dot(v1.get_value(), m.get_value()) + v2_orig)
 
+    def test_gemv_broadcast(self):
+        ''' test gemv with some broadcasted input '''
+        rng = numpy.random.RandomState(unittest_tools.fetch_seed())
+        v1 = theano.shared(numpy.array(rng.uniform(size=(2,)),
+                                       dtype='float32'))
+        v2_orig = numpy.array(rng.uniform(size=(1,)), dtype='float32')
+        v2 = theano.shared(v2_orig)
+        m = theano.shared(numpy.array(rng.uniform(size=(1, 2)),
+                                      dtype='float32'),
+                          broadcastable=(True, False))
+        o = theano.dot(m, v1)
+        f = theano.function([], o + v2, mode=mode_blas_opt)
+
+        # Assert they produce the same output
+        assert numpy.allclose(
+            f(),
+            numpy.dot(m.get_value(), v1.get_value()) + v2.get_value())
+        topo = f.maker.fgraph.toposort()
+        assert sum(isinstance(node.op, Gemv) for node in topo) == 1
+
+        # call gemv directly for mixed broadcast pattern.
+        o = theano.tensor.blas.gemv_no_inplace(v2, 0.5, m, v1, 0.25)
+        f = theano.function([], o, mode=mode_blas_opt)
+        assert numpy.allclose(
+            f(),
+            0.5*numpy.dot(m.get_value(), v1.get_value()) + 0.25*v2.get_value())
+        topo = f.maker.fgraph.toposort()
+        assert sum(isinstance(node.op, Gemv) for node in topo) == 1
+
     def test_gemv_dimensions(self):
         A = T.matrix('A')
         x, y = T.vectors('x', 'y')
@@ -1475,14 +1504,14 @@ class BaseGemv(object):
         x_v = x_v.astype("float32")
         y_v = y_v.astype("float32")
 
-        alpha = T.dscalar('a')
-        a = T.fmatrix('w')
-        x = T.fvector('v')
-        y = T.fvector('t')
+        alpha = T.dscalar('alpha')
+        a = self.shared(a_v)
+        x = self.shared(x_v)
+        y = self.shared(y_v)
 
         rval = T.dot(a, x) * alpha + y
 
-        f = theano.function([a, x, y, alpha], rval, mode=self.mode)
+        f = theano.function([alpha], rval, mode=self.mode)
         # this function is currently optimized so that the gemv is
         # done inplace on a temporarily allocated-buffer, which is
         # then scaled by alpha and to t with a fused elemwise.
@@ -1494,7 +1523,7 @@ class BaseGemv(object):
                 assert node.outputs[0].dtype == 'float32'
         assert n_gemvs == 1, n_gemvs
         self.assertFunctionContains1(f, self.gemv_inplace)
-        f(a_v, x_v, y_v, alpha_v)
+        f(alpha_v)
 
 
 class TestSgemv(TestCase, BaseGemv, unittest_tools.TestOptimizationMixin):
@@ -1508,16 +1537,16 @@ class TestDgemv(TestCase, BaseGemv, unittest_tools.TestOptimizationMixin):
     gemv = theano.tensor.blas.gemv_no_inplace
     gemv_inplace = theano.tensor.blas.gemv_inplace
 
-#The optimization to put Gemv don't work for complex type for now.
+# The optimization to put Gemv don't work for complex type for now.
 # See ticket 653.
-#class TestCgemv(TestCase, BaseGemv):
+# class TestCgemv(TestCase, BaseGemv):
 #    dtype = complex64
 
-#class TestZgemv(TestCase, BaseGemv):
+# class TestZgemv(TestCase, BaseGemv):
 #    dtype = complex128
 
 ###############################################################################
-## Tests for Ger
+# Tests for Ger
 ###############################################################################
 
 
@@ -1549,13 +1578,13 @@ class TestGer_make_node(TestCase):
         self.za = T.zscalar()
 
     def test_works_on_all_valid_dtypes(self):
-        self.assertEquals(self.fm.type,
+        self.assertEqual(self.fm.type,
             ger(self.fm, self.fa, self.fv, self.fv_2).type)
-        self.assertEquals(self.fm.type,
+        self.assertEqual(self.fm.type,
             ger(self.fm, self.fa, self.fv, self.fv_2).type)
-        self.assertEquals(self.fm.type,
+        self.assertEqual(self.fm.type,
             ger(self.fm, self.fa, self.fv, self.fv_2).type)
-        self.assertEquals(self.fm.type,
+        self.assertEqual(self.fm.type,
             ger(self.fm, self.fa, self.fv, self.fv_2).type)
 
     def test_fails_on_invalid_dtypes(self):
@@ -1571,7 +1600,7 @@ class TestGer_make_node(TestCase):
         self.assertRaises(TypeError,
                 ger, self.fm, self.fv1, self.fv, self.fv_2)
         # actually doing the aforementioned dimshuffle makes it work
-        self.assertEquals(self.fm.type,
+        self.assertEqual(self.fm.type,
                 ger(self.fm, self.fv1.dimshuffle(), self.fv, self.fv_2).type)
 
     def test_fails_for_nonmatrix_A(self):
@@ -1783,8 +1812,8 @@ class TestBlasStrides(TestCase):
 
         f_nn = theano.function([], [], updates=[(a, tensor.dot(b, c))],
                 mode=self.mode)
-        #print 'class name:', self.__class__.__name__
-        #theano.printing.debugprint(f_nn)
+        # print 'class name:', self.__class__.__name__
+        # theano.printing.debugprint(f_nn)
         f_nt = theano.function([], [], updates=[(a, tensor.dot(b, c_t.T))],
                 mode=self.mode)
         f_tn = theano.function([], [], updates=[(a, tensor.dot(b_t.T, c))],
@@ -2130,3 +2159,62 @@ class TestBlasStrides(TestCase):
         self.cmp_ger((0, 1), 0, 1)
         self.cmp_ger((1, 0), 1, 0)
         self.cmp_ger((0, 0), 0, 0)
+
+
+class test_infer_shape(unittest_tools.InferShapeTester):
+    def test_dot22(self):
+        x, y = T.matrices('xy')
+        self._compile_and_check(
+            [x, y], [T.blas._dot22(x, y)],
+            [numpy.random.random((2, 3)).astype(config.floatX),
+             numpy.random.random((3, 4)).astype(config.floatX)],
+            T.blas.Dot22)
+
+    def test_dot22scalar(self):
+        x, y = T.matrices('xy')
+        a = T.scalar('a')
+        self._compile_and_check(
+            [x, y, a], [T.blas._dot22scalar(x, y, a)],
+            [numpy.random.random((2, 3)).astype(config.floatX),
+             numpy.random.random((3, 4)).astype(config.floatX),
+             numpy.asarray(0.5, dtype=config.floatX)],
+            T.blas.Dot22Scalar)
+
+    def test_gemm(self):
+        x, y, z = T.matrices('xyz')
+        a = T.scalar('a')
+        b = T.scalar('b')
+        self._compile_and_check(
+            [x, y, a, z, b], [T.blas.gemm(z, a, x, y, b)],
+            [numpy.random.random((2, 3)).astype(config.floatX),
+             numpy.random.random((3, 4)).astype(config.floatX),
+             numpy.asarray(0.5, dtype=config.floatX),
+             numpy.random.random((2, 4)).astype(config.floatX),
+             numpy.asarray(0.5, dtype=config.floatX)],
+            T.blas.Gemm)
+
+    def test_gemv(self):
+        A = T.matrix('A')
+        x, y = T.vectors('xy')
+        a = T.scalar('a')
+        b = T.scalar('b')
+        self._compile_and_check(
+            [y, a, A, x, b], [T.blas.gemv(y, a, A, x, b)],
+            [numpy.random.random((2,)).astype(config.floatX),
+             numpy.asarray(0.5, dtype=config.floatX),
+             numpy.random.random((2, 3)).astype(config.floatX),
+             numpy.random.random((3,)).astype(config.floatX),
+             numpy.asarray(0.5, dtype=config.floatX)],
+            T.blas.Gemv)
+
+    def test_ger(self):
+        A = T.matrix('A')
+        x, y = T.vectors('xy')
+        a = T.scalar('a')
+        self._compile_and_check(
+            [A, a, x, y], [T.blas.ger(A, a, x, y)],
+            [numpy.random.random((2, 3)).astype(config.floatX),
+             numpy.asarray(0.5, dtype=config.floatX),
+             numpy.random.random((2,)).astype(config.floatX),
+             numpy.random.random((3,)).astype(config.floatX)],
+            T.blas.Ger)
