@@ -3,9 +3,9 @@ Test config options.
 """
 
 import unittest
-from theano import config
 from theano.configparser import AddConfigVar, ConfigParam, THEANO_FLAGS_DICT
 
+
 class T_config(unittest.TestCase):
 
     def test_invalid_default(self):
@@ -30,17 +30,15 @@ class T_config(unittest.TestCase):
         except ValueError:
             pass
 
-        try:
-            THEANO_FLAGS_DICT['T_config.test_invalid_default_b'] = 'ok'
-            # This should succeed since we defined a proper value, even
-            # though the default was invalid.
-            AddConfigVar(
-                    'T_config.test_invalid_default_b',
-                    doc='unittest',
-                    configparam=ConfigParam('invalid', filter=filter),
-                    in_c_key=False)
-        finally:
-            # Dicionary clean-up.
-            del THEANO_FLAGS_DICT['T_config.test_invalid_default_b']
+        THEANO_FLAGS_DICT['T_config.test_invalid_default_b'] = 'ok'
+        # This should succeed since we defined a proper value, even
+        # though the default was invalid.
+        AddConfigVar('T_config.test_invalid_default_b',
+                     doc='unittest',
+                     configparam=ConfigParam('invalid', filter=filter),
+                     in_c_key=False)
+
+        # Check that the flag has been removed
+        assert 'T_config.test_invalid_default_b' not in THEANO_FLAGS_DICT
 
         # TODO We should remove these dummy options on test exit.
